#include "hmbdc/Copyright.hpp"
#pragma once
#include "hmbdc/app/Message.hpp"
#include "hmbdc/pattern/GuardedSingleton.hpp"

#include  <functional>
#include  <tuple>

namespace hmbdc { namespace app { namespace utils {

namespace netportal_detail {
template <typename ...RecvMessages>
uint32_t itemSizePower2Num() {
    uint32_t res = 1ul;
    auto valueTypeSize = max_size_in_tuple<std::tuple<RecvMessages...>>::value
        + sizeof(uint64_t);
    while ((1ul << res) < valueTypeSize) res++;
    return res;
}
} //detail

/**
 * @brief an utility that gives finer control to the user that do
 * not rely on hmbdc Contexts
 * @details still needs to use NetContext singleton which needs 
 * to be created first
 * 
 * @tparam NetContext - on the hmbdc net transport Context such as
 * tcpcast::NetContext, ... etc
 * @tparam SendMessageTuple - std tuple containing the message types to 
 * send out - can be empty
 * @tparam RecvMessageTuple - std tuple containing the message types to 
 * receive - default to be empty
 * 
 */
template <typename NetContext, typename SendMessageTuple
    , typename RecvMessageTuple = std::tuple<>>
struct NetPortal;

template <typename NetContext, MessageC ... SendMessages, MessageC ... RecvMessages>
struct NetPortal<NetContext, std::tuple<SendMessages...>, std::tuple<RecvMessages...>> { 
    /**
     * @brief ctor - only can be called after the NetContext singleton is created
     * @details it does not go thru the state machine for sender / receiver
     * hand shake if the netowork transport dictates that
     * calling rotate several times will make it happen
     * 
     * @param cfg a json configure that has a tx and / or rx sections
     * the supported params are the same as individual transport
     */    
    NetPortal(Config cfg)
    : outOneBuffer_(netportal_detail::itemSizePower2Num<RecvMessages...>())
    , sendEng_(sizeof...(SendMessages)?NetContext::instance().createSendTransportEngine(
        (cfg.put("connKeyCheck", false), cfg.resetSection("tx", false))
        , max_size_in_tuple<std::tuple<SendMessages...>>::value)
        : nullptr)
    , recvEng_(sizeof...(RecvMessages)?NetContext::instance().createRecvTransportEngine(
        cfg.resetSection("rx", false)
        , outOneBuffer_)
        : nullptr)
    {}

    /**
     * @brief run the portal to send and receive messages
     * @details it does NOT block - it is not garanteed the user
     * will have sent or received messages. It is expected to be called
     * multiple times
     * thread safe
     */
    void rotate() {
        if (hmbdc_likely(recvEng_)) recvEng_->rotate();
        if (hmbdc_likely(sendEng_)) sendEng_->rotate();
    }

    /**
     * @brief get the incoming messages and dispatch them to 
     * the receiving code
     * @details it does NOT block - it is not garanteed the user
     * will have received messages if there is nothing received. 
     * It is expected to be called whenever the user code feels like receiving messages.
     * This is NOT a thread safe or reentrant function.
     * 
     * @tparam MesssageRecver a type that has the handleMessageCb()
     * callbacks - see app::Client's doc for those callbacks
     * @tparam MesssageRecverInterestsTuple tuple containing the message types
     * of the receiver can handle; If a messages that the receiver is not interested
     * is received, it is simply dropped.
     * @param recver an instance of the receiver
     * @return the message count that has been dispatched to the recver
     * in this call
     */
    template <typename MesssageRecver
        , typename MesssageRecverInterestsTuple = std::tuple<RecvMessages...>>
    size_t pullIn(MesssageRecver&& recver) {
        outOneBuffer_.dispCount = 0;
        if (hmbdc_likely(recvEng_)) {
            outOneBuffer_.postCommit = [&recver](MessageHead& h) {
                using M = typename std::decay<MesssageRecver>::type;
                MessageDispacher<M, MesssageRecverInterestsTuple> disp;
                return disp(recver, h);
            };
            recvEng_->rotate();
            outOneBuffer_.postCommit 
                = [](MessageHead&){return true;}; //clear
        }
        return outOneBuffer_.dispCount;
    }

    /**
     * @brief just run the portal part that send messages
     * @details it does NOT block - it is not garanteed the user
     * will have sent messages depending on underly network transport. 
     * It is expected to be called multiple times
     * thread safe
     * 
     * @details in low traffic cases the messages are always sent out
     * by single call
     */
    void flushOut() {
        if (hmbdc_likely(sendEng_)) sendEng_->rotate();
    }

    /**
     * @brief get a sender for a topic
     * @details thread safe but a slow call 
     * - so the user cashing the result is expected
     * 
     * @param t topic for sending
     * @return sender pointer. The sender is owned by the lib - do not delete
     */
    typename NetContext::Sender* getSender(comm::Topic const& t) {
        if (sendEng_ && sendEng_->match(t))
            return NetContext::instance().getSender(t);
        return nullptr;
    }

    /**
     * @brief get a send engineng
     * @return sender engine pointer. The sender engine is owned by the lib - do not delete
     */
    auto getSendEngine() {
        return sendEng_;
    }

    /**
     * @brief listen to a topic - has impact on the
     * pullIn call
     * @details threadsafe
     * 
     * @param t topic to listen to
     */
    void listenTo(comm::Topic const& t) {
        if (recvEng_) recvEng_->listenTo(t);
    }

    /**
     * @brief stop listening to a topic - has impact on the
     * pullIn call
     * @details threadsafe
     * 
     * @param t topic to forget
     */
    void stopListenTo(comm::Topic const& t) {
        if (recvEng_) recvEng_->stopListenTo(t);
    }

    /**
     * @brief the sending session that are active
     * which is how many receiver are active
     * @details only supported for reliable transport types
     * such as tcpcast, rmast, and rnemap
     * @return the count - return numeric_limits<size_t>::max()
     * when the first receiver is yet to connect (initial)
     */
    size_t sendSessionsRemainingActive() const {
        return sendEng_?sendEng_->sessionsRemainingActive():0;
    }

    /**
     * @brief the recving session that are active
     * which is how many sending parties are active
     * @details only supported for reliable transport types
     * such as tcpcast, rmast, and rnemap
     * @return the count
     */
    size_t recvSessionsRemainingActive() const {
        return recvEng_?recvEng_->sessionsRemainingActive():0;
    }
        
private:
    struct OneBuffer {
        using iterator = hmbdc::pattern::lf_misc::iterator<uint64_t>;
        using value_type = void *;
        const size_t valueTypeSize;
        hmbdc::pattern::lf_misc::chunk_base_ptr<uint64_t> buffer;
        std::function<bool (MessageHead&)> postCommit 
            = [](MessageHead&){return true;};
        size_t dispCount = 0;

        OneBuffer(size_t valueTypeSizePower2Num)
        : valueTypeSize(1ul << valueTypeSizePower2Num)
        , buffer(valueTypeSizePower2Num, 0) {}
        
        size_t capacity() const {return 1;};
        size_t maxItemSize() const {return valueTypeSize;}
        void put(void* item, size_t sizeHint = 0) {
            if (postCommit(*(MessageHead*)(item))) 
                dispCount++;
        }

        template <typename T> void put(T& item) {put(&item, sizeof(T));}
        template <typename T> void putSome(T& item) {put(&item, std::min(sizeof(item), maxItemSize()));}
        iterator claim() {
            return iterator(buffer, 0);
        }
        void commit(iterator it) {
            if (postCommit(*(MessageHead*)(*it)))
                dispCount++;
        }
    };

    OneBuffer outOneBuffer_;
    using SendTransportEngine = typename NetContext::SendTransportEngine;
    SendTransportEngine* sendEng_;
    using RecvTransportEngine = typename NetContext::template RecvTransportEngine<OneBuffer>;
    RecvTransportEngine* recvEng_;
};

}}}

